@@ -17,6 +17,7 @@ from typing import (
 
 from loguru import logger
 from tqdm.auto import tqdm
+import importlib
 
 
 def str2bool(x):
@@ -160,6 +161,31 @@ def flatten_dict(odict: dict, result: Optional[dict] = None, prefix: str = ""):
     return result
 
 
+def flatten_list(lst: list) -> list:
+    """Flatten nested list, anything that is instance of a list get flatten"""
+    output = []
+    for item in lst:
+        if isinstance(item, list):
+            for subitem in item:
+                if isinstance(subitem, list):
+                    output += flatten_list(subitem)
+                else:
+                    output.append(subitem)
+        else:
+            output.append(item)
+    return output
+
+
+def group_by(lst: list, key: Callable) -> dict:
+    odict = {}
+    for item in lst:
+        k = key(item)
+        if k not in odict:
+            odict[k] = []
+        odict[k].append(item)
+    return odict
+
+
 class ParallelMapFnWrapper:
     def __init__(self, fn, ignore_error=False):
         self.fn = fn
@@ -266,3 +292,35 @@ class DictProxy(Dict[K, V2]):
 
     def items(self):
         return ((k, self.access(v)) for k, v in self.odict.items())
+
+
+def import_func(func_ident: str) -> Callable:
+    """Import function from string, e.g., sm.misc.funcs.import_func"""
+    lst = func_ident.rsplit(".", 2)
+    if len(lst) == 2:
+        module, func = lst
+        cls = None
+    else:
+        module, cls, func = lst
+        try:
+            importlib.import_module(module + "." + cls)
+            module = module + "." + cls
+            cls = None
+        except ModuleNotFoundError:
+            pass
+
+    module = importlib.import_module(module)
+    if cls is not None:
+        module = getattr(module, cls)
+
+    return getattr(module, func)
+
+
+class Proxy:
+    """Proxy object, can be used for delayed initialization"""
+
+    def __init__(self, _object):
+        self._object = _object
+
+    def __getattr__(self, name):
+        return getattr(self._object(), name)
